/*
    Copyright 2003 Nasif Akand (nasif@yifan.net)
    http://go.to/zipamp
    http://zipamp.virtualave.net

    This file is part of ZipAmp MP3 software.

    ZipAmp is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    ZipAmp is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this code; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

//This module is the implementation of FAT
//Nasif Akand

#include "zfat.h"

//Following includes aren't required if using CodeVision AVR C
//CodeVision will include automatically if .H files are in project list.
//#include "ztype.h"
//#inlcude "zata.h"
//#inlcude "zlcd.h"
//#include "zcontrol.h"


unsigned long clust2LBA(unsigned long clust) {
//Convert cluster address to LBA address, uses the formula
	return ((clust-2) * Sectors_Per_Cluster) + FirstDataSector;
}

/*
unsigned long LBA2Cluster(unsigned long LBA) {
 return ((LBA-FirstDataSector)/Sectors_Per_Cluster)+2;
}
*/

void InitFAT(){
 	struct partRec *pr;
	struct bootSec710 *boot;
	struct driveID *driveInfo;
	word  rootDirSectors;
	struct FAT32fsInfo *fsInfo; 	
	dword startSector;
	dword FATSz, dataSec, totalSector;
	byte	i;	
        RegWrite(DrvHeadReg,LBAMaster);
	RegWrite(CmdStsReg,0xEC); //ATA identify command.	
	while ((RegRead(CmdStsReg) & SR_DRQ)!=SR_DRQ);  //no error handling
	WriteAddr(IOReg);
	ATA_Read(128);
	driveInfo = (struct driveID *) sectorBuffer.data;

        LCDclrscr();
	for (i=0; i<20; i+=2) {writechar(driveInfo->model[i+1]); writechar(driveInfo->model[i]); }; 
	//Determine drive size. Multiplication and division is done by shifts
	startSector = driveInfo->totalSecHi;
	startSector = ((startSector<<16) + driveInfo->totalSecLo)>>11;	
	gotoxy(1,0);
	writeNumber(startSector); //Display drive size
	writestring("MB ");	
	
	
        //skipping boot code in first 445 bytes. So go to offset 446
        GoToSectorOffset(0,446);
	ATA_Read_Sectors_LBA(0, 32); 	  	//reading 64 byte partition records of 4 partition

  	readPos=0;				//New sector needs to be read	
  	sectorPos=0;
	pr = (struct partRec *) sectorBuffer.data;
	startSector = pr->startLBA;
	ATA_Read_Sectors_LBA(startSector, 64);  //Go to start sector of the partition

	boot = (struct bootSec710 *) sectorBuffer.data; //Load BPB struct pointer

	//All of the following calculations are from Microsoft FAT32 White Paper	
	Sectors_Per_Cluster = boot->BPB.secPerClust;
	Bytes_Per_Sector = boot->BPB.bytesPerSec;

	FirstFATSector = boot->BPB.resSectors + startSector;
	
	rootDirSectors=(((boot->BPB.rootDirEnts * 32)+ (Bytes_Per_Sector - 1)) / Bytes_Per_Sector);
	FATSz = boot->BPB.FATSz16;
	if (FATSz==0) FATSz = boot->BPB.FATSz32;
	if (boot->BPB.totSec16 != 0) totalSector=boot->BPB.totSec16;
	else totalSector=boot->BPB.totSec32;
	dataSec = totalSector - (boot->BPB.resSectors + (boot->BPB.FATs * FATSz) + rootDirSectors);

	FirstDataSector = boot->BPB.resSectors + (boot->BPB.FATs * FATSz) + rootDirSectors + startSector;

	if ((dataSec / Sectors_Per_Cluster) < 65525) {
	 Halt(3); //System is FAT16, not supported in this version.
	} 
	else {
	 FirstDirSector = boot->BPB.rootClust;
	 writestring("FAT32");//FATType=32;   
	} 
	
        GoToSectorOffset(startSector+1,484);	//Goto offset 484 of FSInfo sector (2nd sector)
 	ATA_Read(28);            
	fsInfo= (struct FAT32fsInfo *) sectorBuffer.data;
        DriveSize=fsInfo->freeCluster; //driveSize in # of cluster
					//Next freeCluster address changes everytime new file added or removed
	                                //Thus ZipAmp will see that user modified total files in drive, so
	                                //re-count of all files are needed.

 
	gotoxy(2,0);
	writeNumber((DriveSize*Sectors_Per_Cluster)>>11); //Display approx. free space in MB
	writestring("MB Free");		                  //Right shift of 11 is found from equation given below
							  //Display will give wrong value if free space > 65GB
							  //Because writeNumber function expects word value.
	
	//DriveSize is in # of cluster
	//size = (DriveSize * Sectors_Per_Cluster * Bytes_Per_Sector / (1024 * 1024)) MB
	//Assume Bytes_Per_Sector = 512,
	//Therefore, operation of ( 512 / ( 1024 * 1024) ) = Right shift of 11
 	
}

unsigned char getDirEntry(byte onlyMP3){
//gets a filename.ext, filesize, and start cluster of that file from the directory entry table.
//Returns 1 if file is found, returns 0 if no file found and reaches end of dir table.
//Finds only MP3 files and enters sub-directory if onlyMP3 is set, if its 0 then only reads root dir.
//Each file record in dir is 32 byte long.


	unsigned long cluster;
	word fileNum=0;
	dword extension;
	byte i,j,start,longNameFound;	
	struct direntry *file;	
	struct winentry *longFile;
	
	readPos=0; 

        if ((currentFileNum & dirReadOffset.fileNum) ==0) { //if reading from start of root dir
	  cluster=FirstDirSector;
	  sectorPos=0;
	  subDirLevel=0;   
	  
	  dirString[0]=2;
	  dirString[1]='/';
	  
	}
	else { 
       	 cluster=dirReadOffset.cluster;		//Start from last read cluster position.		
	 sectorPos=dirReadOffset.sectorPos;     //Last read sectorPos
	 fileNum=dirReadOffset.fileNum;		//Last read file number

	 //Find the previous file record read offset in the cluster
 	 GoToSectorOffset(clust2LBA(cluster)+sectorPos,dirReadOffset.offset);
	}               
	

	CLEAR:				//Jump position if an entry was invalid entry, clear fileName
	
	for (i=0; i<40; i++) fileName[i]=0; //Clear file name so all position is NULL.
	extension=0;
	longNameFound=0;

      do { //Directory read loop (infinite, unless a file found or reached end of directory)
      
 	  ATA_Read_Sectors_LBA(clust2LBA(cluster)+sectorPos, 16); //reading 32 byte in each iteration
          if (sectorPos==Sectors_Per_Cluster) {
           currentCluster=cluster;
           cluster=findNextCluster();           
           sectorPos=0; readPos=0;             
          } 	  

 	 
 	 
 	longFile=(struct winentry *) sectorBuffer.data; //Initialize pointer for long file name structure
 	file = (struct direntry *) sectorBuffer.data;   //Initialize pointer for short file name structure
 	
 	if (sectorBuffer.data[0] == 0xE5) {
	  //Deleted entry
 	 goto CLEAR;
 	}

 	 //------------------------------------------------------------------------------        	  
         /*End of a directory  block */ 
 	 if (sectorBuffer.data[0] == 0) {  //We reached end of a dir
 	  if (subDirLevel>0) { //Check if this end of a sub-dir, if true then go to one level up directory
  	    subDirLevel--;
  	    popDir(); //pop cluster address from stackfile to currentCluster
  	    cluster=currentCluster; //set dir cluster to popped cluster
  	    GoToSectorOffset(clust2LBA(cluster)+sectorPos,readPos); //go to dir file entry offset
  	    
  	    //Remove subdir text in dir string
  	    j=dirString[0]-2;
  	    while ((j>1)&&(dirString[j]!='/')) j--;
  	    dirString[0]=j+1;
  	    
  	    goto CLEAR; 	   
 	  }                             

 	  //If we are here then we reached end of root directory.
 	  currentFileNum=0;
 	  return 0; //Done. Finished reading all entries so return 0
 	 } 	 
 	
 	//-------------------------------------------------------------------------------              
	/* Active long name sub-component block. */ 	
 	 if ((longFile->Attr & ATTR_LONG_NAME_MASK) == ATTR_LONG_NAME) {
	   	
 	   start = 13 *( (longFile->Cnt-1) & 0x0F); //Fine offset of text in long file name
 	   
 	   //Following while loops are shortened with EEPROM table longNameOffset. This reduces code size.
           /*i=0;
  	   while ((start<40)&&(i<5)) { fileName[start++]=longFile->Part1[i++]; }
  	   i=0;                                                                 
       	   while ((start<40)&&(i<6)) { fileName[start++]=longFile->Part2[i++]; }
       	   i=0;
  	   while ((start<40)&&(i<2)) { fileName[start++]=longFile->Part3[i++]; }*/
  	   
  	   //We only store text in fileName array if total length of text is less than 40 bytes
  	   i=0;
  	   while ((start<40)&&(i<13)) { fileName[start++] = sectorBuffer.data[longNameOffset[i++]]; }
  	   longNameFound=1;
	 }                                                                        
	 
 	 else {
 	  //Found Short file entry, now determine its type
 	  //------------------------------------------------------------------------------
          for (i=0; i<8; i++) {                               
           //Check for file name validity. There can be garbage in dir entry, that will cause system failure
           //j<47 comes from # of eeprom invalidFileChars table
           for (j=0; j<47; j++) if (invalidFileChars[j]== file->name[i]) goto CLEAR;
           
           //This may look redundant but its done so that we don't have to write another loop to read fileName
	   if (longNameFound==0) fileName[i]=file->name[i]; // Reading file name if long name not found
	  } 
        
  	  /* Found a file. -----------------------------------*/
 	  if ((file->attributes & 0x18 /*(ATTR_DIRECTORY | ATTR_VOLUME_ID)*/ ) == 0x00) {
	
	   	  
           for(i=0;i<3; i++) { //Collect the file extension
            fileExtension[i]=file->extension[i];

           }      
           extension = (*(dword *)fileExtension) & 0x00FFFFFF;
	   // if we wanted MP3 but extesion wasn't MP3 then goto CLEAR
           if ((onlyMP3==1)&&(extension!=0x0033504D)) goto CLEAR; //extension = 03PM (little endian struct)

	   
           fileNum++;						// File found, so increment fileNum	  
	   if (fileNum>currentFileNum) {	             	// Making sure that this entry is not read before  
	     currentFileNum=fileNum;				// New entry, so increment current file number.             	   	  	    	   	     
	     FileSize = file->fileSize;
	     currentCluster = file->highClust;
	     currentCluster = (currentCluster<<16) + file->startCluster;		     
             dirReadOffset.cluster=cluster;
             dirReadOffset.sectorPos=sectorPos;
	     dirReadOffset.fileNum=fileNum;
	     dirReadOffset.offset=readPos; 
	     return 1;
	   }
	   goto CLEAR;	//This file has already been encountered so clear the info

	  }                                                                                       
	  
	  //-------------------------------------------------------------------------------	  	  
          //Directory block
	  if ((file->attributes & 0x18) == ATTR_DIRECTORY) {
  	  /* Found a directory. */ 
	   
  	   if ((onlyMP3==1) && (subDirLevel<8)) { //Enter sub dir only if looking for MP3
  	    currentCluster=cluster; //pushDir will use value of currentCluster to push onto stack
  	    cluster = file->highClust; //Get the sub-directory cluster address
  	    cluster = (cluster<<16) + file->startCluster;  	   
  	    pushDir(); //push return cluster address on stack (gets value from currentCluster)
     	    subDirLevel++;
  	    readPos=0;	//A new cluster has to be read, so set readPos=0
  	    sectorPos=0;

	    //Fix dirString for LCD path display  	    
  	    i=0;
  	    j=dirString[0];
  	    while ((j<20)&&(fileName[i]!=0)) dirString[j++]=fileName[i++];
  	    dirString[j++]='/';  	     
  	    dirString[0]=j;
  	   }
  	   goto CLEAR;		   
	  }	  
	 } 	  
       } 
       while (1); 	 
}


dword findNextCluster(){
 unsigned long FatOffset, sector;
 unsigned int offset;                  

   FatOffset = (currentCluster << 2); //Lshift by 2 will be 1 in FAT16, not supported here
   sector = FirstFATSector + (FatOffset / Bytes_Per_Sector);
   offset = FatOffset % Bytes_Per_Sector;
   readPos=0; sectorPos=0;
   while (readPos<offset) ATA_Read_Sectors_LBA(sector,1);
   ATA_Read_Sectors_LBA(sector+sectorPos,2);//Have to read 1 word in FAT16, not supported here
   return (*(unsigned long *) sectorBuffer.data) & 0x0FFFFFFF;   //FAT32 address return
  
}

void GoToSectorOffset(dword LBASector,word offset) {
  //Find the offset in the sector
   readPos=0;
   while (readPos<offset) ATA_Read_Sectors_LBA(LBASector,1); //go to offset
}

void pushDir() {
 //pushes the directory return address on stack.
 //Also stores global variables sectorPos and readPos
 //Each record is 14 byte long. 
 //Dir return cluster address = 4 bytes
 //                   ReadPos = 2 bytes
 //                 SectorPos = 1 byte
 //                     Total = 7 bytes
 //Due to ZipAmp hardware, we can only store even bytes
 //Thus, we rearrange the 7 byte data in 14 byte space, with every even byte having actual data.
 
 byte offset, i;
 
 
 offset = subDirLevel * 14; //this is storage byte offset, position to store this dir info
 for (i=0; i<4; i++) //Arrange the cluster address data on even byte postion
  sectorBuffer.data[offset + (i<<1)]=currentCluster>>(24-(i<<3)); //1 word
  

 sectorBuffer.data[offset+8]=readPos>>8; //arrange readPos on even byte
 sectorBuffer.data[offset+10]=readPos;
 
 sectorBuffer.data[offset+12]=sectorPos;

 offset=offset>>1; //divide by 2 to get word offset
 if (offset>0) { //Go to offset if its not 0th position in sector
  readPos=0;             
  ATA_Read_Sectors_LBA(dirStackLBA, offset ); //Load previous dir info in buffer
 }                                           
 
 ATA_WriteSector(dirStackLBA); //Now write the buffer into file ZIPAMP.SYS pointed by dirStackLBA.
}  

void popDir(){ //Pops directory info from stack, reverse process of pushDir.
 byte offset,i;
 offset=subDirLevel*14; //Offset we have to read now
 readPos=0;
 ATA_Read_Sectors_LBA(dirStackLBA, 128); //Load whole dir stack
 
 //Arrange the data
 for (i=0; i<7; i+=2) currentCluster=(currentCluster<<8) + sectorBuffer.data[offset+i];

 readPos=sectorBuffer.data[offset+8];
 readPos=(readPos<<8) + sectorBuffer.data[offset+10];
 
 sectorPos=sectorBuffer.data[offset+12];

}

void saveConfig() {
//Saves player status on ZIPAMP.CFG file.

 struct configuration *cfg;
 byte i;
 cfg = (struct configuration *) sectorBuffer.data;
 cfg->Sign1=0xA5;			//DataValid signature
 cfg->SoundMode=soundMode;		//PlayType
 cfg->PlayMode=playMode;		//Play mode
 cfg->DriveSize=DriveSize;		//Drive size
 cfg->TotalFiles=totalFiles;		//Total files
 cfg->LastFile=currentFileNum-1;        //Played file
 cfg->Sign2=0x5A;			//DataValid end sign
 
 //Hardware lets us save only even bytes. So fix the array. Position i = position i*2
 //13 comes from total size of struct configuration byte size+1 (12+1 bytes)
 //Position 0 doesn't need to be fixed, its already in correct position.
 for (i=13; i>0; i--) sectorBuffer.data[i*2]=sectorBuffer.data[i]; //Format data for saving, gap every odd byte
 ATA_WriteSector(cfgFileLBA);
}

byte loadConfig() {
//Load player status from ZIPAMP.CFG file.

 struct configuration *cfg;
 byte i;
  
 readPos=0;
 ATA_Read_Sectors_LBA(cfgFileLBA,128);
 for (i=1; i<=13; i++) sectorBuffer.data[i]=sectorBuffer.data[i*2]; //Format data for reading, ungap every odd byte
 cfg = (struct configuration *) sectorBuffer.data;
 if (cfg->DriveSize==DriveSize) { //If hard drive info matches
  totalFiles=cfg->TotalFiles;
  soundMode=cfg->SoundMode;
  playMode=cfg->PlayMode;     
  currentFileNum=cfg->LastFile;  
  return 1; //Load was successful.
 }

 return 0; //Load failed.
}